/*
 * Copyright 2009-2010 Londata Limited
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package net.launchpad.xmlzebra

import collection.Map
import collection.immutable.EmptyMap
import collection.mutable.HashMap
import scala.List
import java.io._
import java.math.BigInteger
import javax.xml.namespace.QName

import java.nio.charset.Charset._
import javax.xml.stream.{XMLStreamWriter, XMLOutputFactory}
import org.slf4j.{Logger,LoggerFactory}

import org.apache.xmlbeans._
import impl.xb.xsdschema.SchemaDocument
import org.w3c.dom.{Document, Attr}

object SchemaPathGenerator {
  val XSP_PREFIX = "xsp"
  val XSP_NS = "http://xml.londata.com/2009/ns/schemaPath"

  val XSD_PREFIX = "xs"
  val XSD_NS = "http://www.w3.org/2001/XMLSchema"

  val XSI_PREFIX = "xsi"
  val XSI_NS = "http://www.w3.org/2001/XMLSchema-instance"

  val XMLNS_FLAG = "-xmlns:"
  
  type QNamed = {def getName(): QName}

  var schemaTypeSystem: SchemaTypeSystem = null
  var outStream: OutputStream = System.out
  var outputFactory = XMLOutputFactory newInstance
  var xmlout: XMLStreamWriter = null
  var schemaFiles = HashMap[XmlObject, File]()

  var namespaces = HashMap[String, String]()

  setNamespacePrefix(XSP_PREFIX, XSP_NS)
  setNamespacePrefix(XSD_PREFIX, XSD_NS)
  setNamespacePrefix(XSI_PREFIX, XSI_NS)

  var nsIndex = 0

  val log = LoggerFactory getLogger(SchemaPathGenerator getClass)

  def usage {
    println("Usage: scala ... com.londata.zebra.SchemaPathGenerator [ -o <outfile> ] { <xml-schema-file-or-dir-path> | -xmlns:<prefix>=<uri> }")
  }

  def main(args: Array[String]) {
    try {
      // Process command-line arguments
      var argIndex = 0
      if (args(argIndex) == "-o") { // process output file specification
        argIndex += 1
        if (argIndex >= args.length) {
          log error "Output file not specificied after '-o' flag"
          usage
          System exit 1
        }

        val outFile = new File(args(argIndex))
        argIndex += 1
        if (outFile.exists && !outFile.canWrite) {
          log error "Cannot write to output file: " + outFile.getAbsolutePath
          usage
          System exit 1
        }
        outStream = new FileOutputStream(outFile)
      }

      while (argIndex < args.length) {
        if (args(argIndex) startsWith XMLNS_FLAG) { // process XML namespace declaration
          val prefixAndUri = args(argIndex) substring XMLNS_FLAG.length
          if (prefixAndUri.length >= 1) {
            if (prefixAndUri contains '=') {
              val equalsIndex = prefixAndUri.indexOf('=')
              setNamespacePrefix(prefixAndUri.substring(0, equalsIndex), prefixAndUri.substring(equalsIndex+1))
            } else {
              setNamespacePrefix(prefixAndUri, "")
            }
          }
        } else { // assume argument is a file or directory
          val file = new File(args(argIndex))
          if (file exists) {
            loadFile(file)
          }
        }
        argIndex += 1
      }

      // Start XML Schema processing
      log info "No. of XML Schemas: " + schemaFiles.size
      xmlout = outputFactory createXMLStreamWriter(outStream, "UTF-8")

      // Get namespace prefixes for the target namespaces; if not found, create one for each
      for (schema <- schemaFiles.keySet) {
        val targetNamespace = schema.asInstanceOf[SchemaDocument].getSchema.getTargetNamespace
        if (namespaces.get(targetNamespace) == None) {
          var nsPrefix = ""
          val document = schema.asInstanceOf[SchemaDocument].getDomNode.asInstanceOf[Document]
          val root = document.getDocumentElement
          val attributes = root.getAttributes
          for (
            attrIndex <- 0 until attributes.getLength
            if nsPrefix == ""
          ) {
            val attribute = attributes.item(attrIndex).asInstanceOf[Attr]
            if (attribute.getValue == targetNamespace) {
              if (attribute.getName.startsWith("xmlns:")) {
                nsPrefix = attribute.getName.substring("xmlns:".length)
                if (nsPrefix == null) {
                  nsPrefix = ""
                }
              }
            }
          }
          if (nsPrefix == "") {
            nsPrefix = "ns" + nsIndex
            nsIndex += 1
          }
          setNamespacePrefix(nsPrefix, targetNamespace)
        }
      }

      schemaTypeSystem = XmlBeans compileXsd (schemaFiles.keySet.toArray, XmlBeans.getBuiltinTypeSystem(), null);

      xmlout writeStartDocument ("UTF-8", "1.0")
      xmlout writeStartElement (XSP_PREFIX, "schemaPaths", XSP_NS)

      xmlout writeNamespace (XSP_PREFIX, XSP_NS)
      xmlout writeNamespace (XSD_PREFIX, XSD_NS)
      xmlout writeNamespace (XSI_PREFIX, XSI_NS)

      addNamespacePrefixes (namespaces)
      addGlobalElementPaths (schemaTypeSystem)

      xmlout writeEndElement () // xsp:schemaPaths
      xmlout writeEndDocument ()
      outStream flush

      log info "Done."
    }
    catch {
      case e: Throwable =>
        val totalMeg = (Runtime.getRuntime().totalMemory() + (512*1024)) / (1024*1024)
        val freeMeg = (Runtime.getRuntime().freeMemory() + (512*1024)) / (1024*1024)
        e.printStackTrace
        println("Memory: total = " + totalMeg + "M, free = " + freeMeg + "M")
    }
  }

  private def setNamespacePrefix(prefix: String, namespaceUri: String) {
    if ((prefix != null) && (prefix.length >= 1)) {
      if (!(namespaces contains namespaceUri)) {
        namespaces put (namespaceUri, prefix)
      }
    }
  }

  private def addNamespacePrefixes(namespaces: Map[String, String]) {
    val uris = namespaces.keySet.toList.sort((x, y) => x < y)
    for (
      uri <- uris
      if uri != XSP_NS
    ) {
      xmlout writeStartElement (XSP_PREFIX, "ns", XSP_NS)
      xmlout writeAttribute ("prefix", namespaces(uri))
      xmlout writeCharacters uri
      xmlout writeEndElement // xsp:ns
    }
  }

  private def addGlobalElementPaths(schemaTypeSystem: SchemaTypeSystem) {
    log info "No. of global elements: " + schemaTypeSystem.globalElements.size
    for (
      globalElement <- schemaTypeSystem.globalElements.toList.sort(compareQNames)
      if !globalElement.isAbstract
    ) yield { // TODO: really needs a way to add an extra filter to select root elements from other global elements
      addLocalElementPath (globalElement, "", List(), new EmptyMap[String, AnyRef]())
    }
  }

  private def addLocalElementPath(
    element: SchemaLocalElement, parentPath: String, complexTypeStack: List[SchemaType], parameters: Map[String, AnyRef]
  ) {
    val elementName = qName(element)
    log debug "element: " + elementName
    val hasSimpleType = (element.getType.isSimpleType) || (element.getType.getContentType == SchemaType.SIMPLE_CONTENT)
    val hasElements = !hasSimpleType && ((element.getType.getContentType == SchemaType.ELEMENT_CONTENT) || (element.getType.getContentType == SchemaType.MIXED_CONTENT))
    val minOccurs = (parameters.get("minOccurs") getOrElse element.getMinOccurs).asInstanceOf[BigInteger]
    val minOccursString = if (minOccurs == null) null else if (minOccurs == BigInteger.ONE) null else minOccurs.toString
    val maxOccurs = (parameters.get("maxOccurs") getOrElse element.getMaxOccurs).asInstanceOf[BigInteger]
    val maxOccursString = if (minOccurs == null) null else if (maxOccurs == BigInteger.ONE) null else if (maxOccurs == null) "unbounded" else maxOccurs.toString
    val path =
      parentPath + "/" + qName(element) +
        (if (hasSimpleType) "[base=" + qName(element.getType.getPrimitiveType) + "]" else "") +
        (if (element.isDefault && !element.isFixed) "[default=" + (element getDefaultText) + "]" else "" ) +
        (if (element.isFixed) "[fixed=" + (element getDefaultText) + "]" else "" ) +
        getMultiplicity(minOccurs, maxOccurs)
    log debug "path: " + path 
    if (!element.isAbstract) { // don't create a path for an abstract element
      if (!element.getType.isAbstract) { // don't create a path for an element with an abstract type
        xmlout writeStartElement (XSP_PREFIX, "element", XSP_NS)
        xmlout writeAttribute ("name", elementName)
        if (minOccursString != null) { xmlout writeAttribute ("minOccurs", minOccursString) }
        if (maxOccursString != null) { xmlout writeAttribute ("maxOccurs", maxOccursString) }
        if (hasSimpleType) {
          if (element.isDefault && !element.isFixed) { xmlout writeAttribute ("default", element getDefaultText) }
          if (element isFixed) { xmlout writeAttribute ("fixed", element getDefaultText) }
          xmlout writeAttribute ("base", qName(element.getType.getPrimitiveType))
        } else {
          if (element.getType.getContentType == SchemaType.MIXED_CONTENT) { xmlout writeAttribute ("mixed", "true") }
        }
        xmlout writeAttribute ("path", path)
        if (hasSimpleType) {
          addSimpleType (element.getType, path)
        }
        addAttributePaths (element, path)
        if (!hasSimpleType && hasElements) {
          if (complexTypeStack contains element.getType) {
            xmlout writeStartElement (XSP_PREFIX, "recursion", XSP_NS)
            xmlout writeAttribute ("type", qName(element.getType))
            xmlout writeAttribute ("path", path + "/...[recursion=" + qName(element.getType) + "]")
            xmlout writeEndElement // xsp:element
          } else {
            addContentModel(element.getType.getContentModel, path, element.getType :: complexTypeStack)
          }
        }
        xmlout writeEndElement // xsp:element
      }

      // Now add entries for extension types, if any
      if (!(element.getType.blockExtension || element.getType.finalExtension)) {
        // Locate extension types
        var extensionTypes = List[SchemaType]()
        for (
          globalType <- schemaTypeSystem.globalTypes
          if isExtensionTypeOf(element.getType, globalType) && !globalType.isAbstract
        ) {
          extensionTypes = globalType :: extensionTypes
        }

        // Add an entry for each extension type
        for (extensionType <- extensionTypes.sort(compareQNames)) {
          log debug "extension type " + qName(extensionType) + " for type " + qName(element.getType)
          val hasSimpleType = (extensionType.isSimpleType) || (extensionType.getContentType == SchemaType.SIMPLE_CONTENT)
          val path = parentPath + "/" + qName(element) +
                  "[xsi:type=" + qName(extensionType) + "]" +
                  (if (hasSimpleType) "[base=" + qName(extensionType.getPrimitiveType) + "]" else "") +
                  (if (element.isDefault && !element.isFixed) "[default=" + (element getDefaultText) + "]" else "" ) +
                  (if (element.isFixed) "[fixed=" + (element getDefaultText) + "]" else "" ) +
                  getMultiplicity(element.getMinOccurs, element.getMaxOccurs)
          log debug "path: " + path
          val hasElements = (extensionType.getContentType == SchemaType.ELEMENT_CONTENT) || (extensionType.getContentType == SchemaType.MIXED_CONTENT)
          xmlout writeStartElement (XSP_PREFIX, "element", XSP_NS)
          xmlout writeAttribute ("name", elementName)
          xmlout writeAttribute ("xsi:type", qName(extensionType))
          if (minOccursString != null) { xmlout writeAttribute ("minOccurs", minOccursString) }
          if (maxOccursString != null) { xmlout writeAttribute ("maxOccurs", maxOccursString) }
          if (hasSimpleType) {
            if (element.isDefault && !element.isFixed) { xmlout writeAttribute ("default", element getDefaultText) }
            if (element isFixed) { xmlout writeAttribute ("fixed", element getDefaultText) }
            xmlout writeAttribute ("base", qName(extensionType.getPrimitiveType))
          } else {
            if (extensionType.getContentType == SchemaType.MIXED_CONTENT) { xmlout writeAttribute ("mixed", "true") }
          }
          xmlout writeAttribute ("path", path)
          addAttributePaths (element, path)
          if (hasElements) {
            if (complexTypeStack contains extensionType) {
              xmlout writeStartElement (XSP_PREFIX, "recursion", XSP_NS)
              xmlout writeAttribute ("type", qName(extensionType))
              xmlout writeEndElement // xsp:element
            } else {
              addContentModel(extensionType.getContentModel, path, extensionType :: complexTypeStack)
            }
          }
          xmlout writeEndElement // xsp:element
        }
      }
    }

    log debug "about to check for substitutions for '" + qName(element) + "'"
    if (element.isInstanceOf[SchemaGlobalElement]) {
      val candidateElement = element.asInstanceOf[SchemaGlobalElement]
      log debug "checking for substitutions for '" + qName(candidateElement) + "'"
      for (
        globalElement <- schemaTypeSystem.globalElements.toList.sort(compareQNames)
        if (globalElement.substitutionGroup == candidateElement) && (!globalElement.isAbstract)
      ) {
        log debug "adding substitution '" + qName(globalElement) + "'"
        addLocalElementPath(globalElement, parentPath, complexTypeStack, collection.immutable.Map("minOccurs" -> minOccurs, "maxOccurs" -> maxOccurs));
      }
    } else { // need to check differently for a local element
      val substitutionQNames = complexTypeStack.first.getElementProperty(element.getName).acceptedNames
      if (substitutionQNames.size >= 1) {
        log debug "checking for substitutions for '" + qName(element) + "'"
        for (
          globalElement <- schemaTypeSystem.globalElements.toList.sort(compareQNames)
          if (substitutionQNames contains globalElement.getName) && (!globalElement.isAbstract)
        ) {
          log debug "adding substitution '" + qName(globalElement) + "'"
          addLocalElementPath(globalElement, parentPath, complexTypeStack, collection.immutable.Map("minOccurs" -> minOccurs, "maxOccurs" -> maxOccurs));
        }
      }
    }
  }

  private def isExtensionTypeOf(baseType: SchemaType, someType: SchemaType): Boolean = {
    if (someType.getBaseType == null) {
      false
    } else if (someType.getBaseType == baseType) {
      true
    } else {
      isExtensionTypeOf(baseType, someType.getBaseType)
    }
  }

  private def addContentModel(contentModel: SchemaParticle, parentPath: String, complexTypeStack: List[SchemaType]) {
    val minOccurs = if (contentModel.getMinOccurs == null) null else if (contentModel.getMinOccurs == BigInteger.ONE) null else contentModel.getMinOccurs.toString
    val maxOccurs = if (contentModel.getMinOccurs == null) null else if (contentModel.getMaxOccurs == BigInteger.ONE) null else if (contentModel.getMaxOccurs == null) "unbounded" else contentModel.getMaxOccurs.toString

    contentModel.getParticleType match {
      case SchemaParticle.SEQUENCE =>
        val multiplicity = getMultiplicity(contentModel.getMinOccurs, contentModel.getMaxOccurs)
        val path = if (multiplicity == "") parentPath else parentPath + "/xs:sequence" + multiplicity
        log debug "path: " + path
        xmlout writeStartElement (XSP_PREFIX, "sequence", XSP_NS)
        if (minOccurs != null) { xmlout writeAttribute ("minOccurs", minOccurs) }
        if (maxOccurs != null) { xmlout writeAttribute ("maxOccurs", maxOccurs) }
        xmlout writeAttribute ("path", path)
        addContentModelChildren (contentModel, path, complexTypeStack)
        xmlout writeEndElement // xsp:sequence
      case SchemaParticle.CHOICE =>
        val multiplicity = getMultiplicity(contentModel.getMinOccurs, contentModel.getMaxOccurs)
        val path = parentPath + "/xs:choice" + multiplicity
        log debug "path: " + path
        xmlout writeStartElement (XSP_PREFIX, "choice", XSP_NS)
        if (minOccurs != null) { xmlout writeAttribute ("minOccurs", minOccurs) }
        if (maxOccurs != null) { xmlout writeAttribute ("maxOccurs", maxOccurs) }
        xmlout writeAttribute ("path", path)
        addContentModelChildren (contentModel, path, complexTypeStack)
        xmlout writeEndElement // xsp:choice
      case SchemaParticle.ALL =>
        val multiplicity = getMultiplicity(contentModel.getMinOccurs, contentModel.getMaxOccurs)
        val path = parentPath + "/xs:all" + multiplicity
        log debug "path: " + path
        xmlout writeStartElement (XSP_PREFIX, "all", XSP_NS)
        if (minOccurs != null) { xmlout writeAttribute ("minOccurs", minOccurs) }
        if (maxOccurs != null) { xmlout writeAttribute ("maxOccurs", maxOccurs) }
        xmlout writeAttribute ("path", path)
        addContentModelChildren (contentModel, path, complexTypeStack)
        xmlout writeEndElement // xsp:all
      case SchemaParticle.WILDCARD =>
        val multiplicity = getMultiplicity(contentModel.getMinOccurs, contentModel.getMaxOccurs)
        val path = parentPath + "/xs:any" + multiplicity
        log debug "path: " + path
        xmlout writeStartElement (XSP_PREFIX, "any", XSP_NS)
        if (minOccurs != null) { xmlout writeAttribute ("minOccurs", minOccurs) }
        if (maxOccurs != null) { xmlout writeAttribute ("maxOccurs", maxOccurs) }
        xmlout writeAttribute ("namespace", "TODO")// TODO
        xmlout writeAttribute ("processContents", "TODO") // TODO
        xmlout writeAttribute ("path", path)
        xmlout writeEndElement // xsp:any
      case SchemaParticle.ELEMENT =>
        addLocalElementPath (contentModel.asInstanceOf[SchemaLocalElement], parentPath, complexTypeStack, new EmptyMap[String, AnyRef]()) // NB: not obvious at all from docs that you can cast an element particle into a local element
      case _ => // ignore, shouldn't occur
    }
  }

  private def addContentModelChildren(contentModel: SchemaParticle, parentPath: String, complexTypeStack: List[SchemaType]) {
    for (child <- contentModel.getParticleChildren) {
      addContentModel (child, parentPath, complexTypeStack)
    }
  }

  private def addAttributePaths(element: SchemaLocalElement, parentPath: String) {
    if ((element.getType.getAttributeModel != null) && (element.getType.getAttributeModel.getAttributes != null)) {
      for (attribute <- element.getType.getAttributeModel.getAttributes.toList.sort(compareQNames)) {
        val attributeName = qName(attribute)
        val use = (
          attribute getUse match {
            case SchemaLocalAttribute.OPTIONAL => "optional"
            case SchemaLocalAttribute.PROHIBITED => "prohibited"
            case SchemaLocalAttribute.REQUIRED => "required"
            case _ => "???" // ignore: should never occur
          }
        )
        val path = parentPath + "/@" + attributeName
        log debug "path: " + path
        xmlout writeStartElement (XSP_PREFIX, "attribute", XSP_NS)
        xmlout writeAttribute ("name", attributeName)
        xmlout writeAttribute ("use", use)
        if (attribute.isDefault && !attribute.isFixed) { xmlout writeAttribute ("default", attribute getDefaultText) }
        if (attribute isFixed) { xmlout writeAttribute ("fixed", attribute getDefaultText) }
        if (attribute.getType.getSimpleVariety == SchemaType.ATOMIC) { xmlout writeAttribute ("base", qName(attribute.getType.getPrimitiveType)) }
        xmlout writeAttribute (
          "path",
          path + "[base=" + qName(attribute.getType.getPrimitiveType) + "]" +
            (if (attribute.isDefault && !attribute.isFixed) "[default=" + (attribute getDefaultText) + "]" else "" ) +
            (if (attribute.isFixed) "[fixed=" + (attribute getDefaultText) + "]" else "" ) +
            (if (attribute.getUse == SchemaLocalAttribute.OPTIONAL) "?" else "")
        )
        addSimpleType (attribute.getType, path)
        xmlout writeEndElement // xsp:attribute
      }
    }
  }

  private def addSimpleType(simpleType: SchemaType, parentPath: String) {
    require((simpleType isSimpleType) || (simpleType.getContentType == SchemaType.SIMPLE_CONTENT))
    simpleType getSimpleVariety match {
      case SchemaType.ATOMIC =>
        addFacets (simpleType, parentPath)
      case SchemaType.LIST =>
        xmlout writeStartElement (XSP_PREFIX, "simpleType", XSP_NS)
        xmlout writeAttribute ("TODO", "LIST")
        xmlout writeAttribute ("path", parentPath + "/xs:simpleType[TODO=LIST]")
        xmlout writeEndElement // xsp:simpleType
      case SchemaType.UNION =>
        xmlout writeStartElement (XSP_PREFIX, "simpleType", XSP_NS)
        xmlout writeAttribute ("TODO", "UNION")
        xmlout writeAttribute ("path", parentPath + "/xs:simpleType[TODO=UNION]")
        xmlout writeEndElement // xsp:simpleType
      case _ => // ignore: should never occur
    }
  }

  private def addFacets(schemaType: SchemaType, parentPath: String) {
    addFacetValue (schemaType, SchemaType.FACET_PATTERN, "pattern", parentPath)
    addFacetValue (schemaType, SchemaType.FACET_MIN_LENGTH, "minLength", parentPath)
    addFacetValue (schemaType, SchemaType.FACET_LENGTH, "length", parentPath)
    addFacetValue (schemaType, SchemaType.FACET_MAX_LENGTH, "maxLength", parentPath)

    if (schemaType.getEnumerationValues != null) {
      for (enumeration <- schemaType.getEnumerationValues.toList.sort((x,y) => x.getStringValue < y.getStringValue)) {
        val enumerationValue = enumeration.getStringValue
        xmlout writeStartElement (XSP_PREFIX, "enumeration", XSP_NS)
        xmlout writeAttribute ("value", enumerationValue)
        xmlout writeAttribute ("path", parentPath + "/xs:enumeration/@value=" + enumerationValue)
        xmlout writeEndElement
      }
    }

    addFacetValue (schemaType, SchemaType.FACET_TOTAL_DIGITS, "totalDigits", parentPath)
    addFacetValue (schemaType, SchemaType.FACET_FRACTION_DIGITS, "fractionDigits", parentPath)
    addFacetValue (schemaType, SchemaType.FACET_MIN_INCLUSIVE, "minInclusive", parentPath)
    addFacetValue (schemaType, SchemaType.FACET_MIN_EXCLUSIVE, "minExclusive", parentPath)
    addFacetValue (schemaType, SchemaType.FACET_MAX_EXCLUSIVE, "maxExclusive", parentPath)
    addFacetValue (schemaType, SchemaType.FACET_MAX_INCLUSIVE, "maxInclusive", parentPath)

    /* [ABC] have decided that whitespace facet isn't sufficiently interesting
    schemaType getWhiteSpaceRule match {
      case SchemaType.WS_COLLAPSE =>
        xmlout writeStartElement (XSP_PREFIX, "whiteSpace", XSP_NS)
        xmlout writeAttribute ("value", "collapse")
        xmlout writeEndElement
      case SchemaType.WS_PRESERVE => // default setting, so don't include it
      case SchemaType.WS_REPLACE =>
        xmlout writeStartElement (XSP_PREFIX, "whiteSpace", XSP_NS)
        xmlout writeAttribute ("value", "replace")
        xmlout writeEndElement
      case _ => // ignore
    }
    */
  }

  private def addFacetValue(schemaType: SchemaType, facetEnum: Int, facetName: String, parentPath: String) {
    // TODO: this code is NOT working for pattern facets.  Don't know why right now.
    getFacetValue(schemaType, facetEnum) match {
      case Some(value) =>
        xmlout writeStartElement (XSP_PREFIX, facetName, XSP_NS)
        xmlout writeAttribute ("value", value)
        xmlout writeAttribute ("path", parentPath + "/xs:" + facetName + "/@value=" + value)
        xmlout writeEndElement
      case None => // ignore
    }
  }

  private def getFacetValue(schemaType: SchemaType, facetEnum: Int): Option[String] = {
    val facet = schemaType.getFacet(facetEnum)
    if (facet == null) {
      None
    } else {
      val facetValue = facet.getStringValue
      if (facetValue == null) None else Some(facetValue)
    }
  }

  private def loadFile(file: File) {
    if (file isDirectory) {
      for (dirFile <- file.listFiles) {
        loadFile(dirFile)
      }
    } else if (isSchema(file)) {
      var xmlObject: Option[XmlObject] = None

      try {
        xmlObject = Some(XmlObject.Factory.parse(file))
      }
      catch {
        case e => xmlObject = None
      }

      xmlObject match {
        case Some(xo: XmlObject) => schemaFiles put(xo, file)
        case _ => // ignore
      }
    }
  }

  private def isSchema(file: File): Boolean = {
    return file.getName.endsWith(".xsd");
  }

  private def compareQNames(x: QNamed, y: QNamed): Boolean = {
    // qName(x) < qName(y)
    curie(x) < curie(y)
  }

  private def qName(xmlDefn: QNamed): String = {
    val name = xmlDefn.getName
    val prefix = namespaces.getOrElse(name.getNamespaceURI, "")
    if (!isEmpty(prefix)) {
      prefix + ":" + name.getLocalPart
    } else if (!isEmpty(name.getNamespaceURI)) {
      "{" + name.getNamespaceURI + "}" + name.getLocalPart
    } else {
      name.getLocalPart
    }
  }

  private def curie(xmlDefn: QNamed): String = {
    val name = xmlDefn.getName
    "{" + name.getNamespaceURI + "}" + name.getLocalPart
  }

  private def isEmpty(s: String): Boolean = {
    (s == null) || (s.length == 0)
  }

  private def getMultiplicity(minOccurs: BigInteger, maxOccurs: BigInteger): String = {
    minOccurs match {
      case null => ""
      case BigInteger.ONE =>
        maxOccurs match {
          case null => "+" // [1..*]
          case BigInteger.ONE => "" // [1..1]
          case _ => "[1.." + maxOccurs + "]"
        }
      case BigInteger.ZERO =>
        maxOccurs match {
          case null => "*" // [0..*]
          case BigInteger.ZERO => "[0]" // [0..0] - shouldn't happen, but needed for debugging
          case BigInteger.ONE => "?" // [0..1]
          case _ => "[0.." + maxOccurs + "]"
        }
      case _ =>
        maxOccurs match {
          case null => "[" + minOccurs + "..*]"
          case _ =>
            if (minOccurs == maxOccurs) "[" + minOccurs + "]"
            else "[" + minOccurs + ".." + maxOccurs + "]"
        }
    }
  }

}